import asyncio
from typing import Callable, Union, List
from collections import UserList

from edsl.jobs.buckets import ModelBuckets
from edsl.questions.QuestionBase import QuestionBase
from edsl.exceptions import InterviewErrorPriorTaskCanceled
from edsl.jobs.tokens.TokenUsage import TokenUsage

from edsl.jobs.tasks.task_status_enum import TaskStatus, TaskStatusDescriptor
from edsl.jobs.tasks.task_management import TokensUsed
from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary

from edsl.jobs.Answers import Answers
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage


from edsl.jobs.tasks.TaskStatusLog import TaskStatusLog


class QuestionTaskCreator(UserList):
    """Class to create and manage question tasks with dependencies.

    It is a UserList with all the tasks that must be completed before the focal task can be run.
    When called, it returns an asyncio.Task that depends on the tasks that must be completed before it can be run.
    """

    task_status = TaskStatusDescriptor()

    def __init__(
        self,
        *,
        question: QuestionBase,
        answer_question_func: Callable,
        model_buckets: ModelBuckets,
        token_estimator: Union[Callable, None] = None,
        iteration: int = 0,
    ):
        super().__init__([])
        self.answer_question_func = answer_question_func
        self.question = question
        self.iteration = iteration

        self.model_buckets = model_buckets
        self.requests_bucket = self.model_buckets.requests_bucket
        self.tokens_bucket = self.model_buckets.tokens_bucket
        self.status_log = TaskStatusLog()

        def fake_token_estimator(question):
            return 1

        self.token_estimator = token_estimator or fake_token_estimator

        # Assume that the task is *not* from the cache until we know otherwise.
        # the _run_focal_task might flip this bit later.
        self.from_cache = False

        self.cached_token_usage = TokenUsage(from_cache=True)
        self.new_token_usage = TokenUsage(from_cache=False)

        self.task_status = TaskStatus.NOT_STARTED

    def add_dependency(self, task: asyncio.Task) -> None:
        """Adds a task dependency to the list of dependencies."""
        self.append(task)

    def __repr__(self):
        return f"QuestionTaskCreator(question = {repr(self.question)})"

    def generate_task(self, debug) -> asyncio.Task:
        """Creates a task that depends on the passed-in dependencies."""
        task = asyncio.create_task(self._run_task_async(debug))
        # TODO: This is a bit hacky.
        task.edsl_name = self.question.question_name
        task.depends_on = [x.edsl_name for x in self]
        return task

    def estimated_tokens(self) -> int:
        """Estimates the number of tokens that will be required to run the focal task."""
        return self.token_estimator(self.question)

    def token_usage(self) -> TokensUsed:
        """Returns the token usage for the task."""
        # return {'cached_tokens': self.cached_token_usage, 'new_tokens': self.new_token_usage}
        return TokensUsed(
            cached_tokens=self.cached_token_usage, new_tokens=self.new_token_usage
        )

    async def _run_focal_task(self, debug) -> Answers:
        """Runs the focal task i.e., the question that we are interested in answering.
        It is only called after all the dependency tasks are completed.
        """

        requested_tokens = self.estimated_tokens()
        if (estimated_wait_time := self.tokens_bucket.wait_time(requested_tokens)) > 0:
            self.task_status = TaskStatus.WAITING_FOR_TOKEN_CAPACITY

        await self.tokens_bucket.get_tokens(requested_tokens)

        if (estimated_wait_time := self.requests_bucket.wait_time(1)) > 0:
            self.waiting = True
            self.task_status = TaskStatus.WAITING_FOR_REQUEST_CAPACITY

        await self.requests_bucket.get_tokens(1)

        self.task_status = TaskStatus.API_CALL_IN_PROGRESS
        try:
            results = await self.answer_question_func(
                question=self.question, debug=debug, task=self
            )
            self.task_status = TaskStatus.SUCCESS
        except Exception as e:
            self.task_status = TaskStatus.FAILED
            raise e

        if "cached_response" in results:
            if results["cached_response"]:
                # Gives back the tokens b/c the API was not called.
                self.tokens_bucket.add_tokens(requested_tokens)
                self.requests_bucket.add_tokens(1)
                self.from_cache = True

        tracker = self.cached_token_usage if self.from_cache else self.new_token_usage

        # TODO: This is hacky. The 'func' call should return an object that definitely has a 'usage' key.

        usage = results.get("usage", {"prompt_tokens": 0, "completion_tokens": 0})
        prompt_tokens = usage.get("prompt_tokens", 0)
        completion_tokens = usage.get("completion_tokens", 0)
        tracker.add_tokens(
            prompt_tokens=prompt_tokens, completion_tokens=completion_tokens
        )
        # self.task_status = TaskStatus.FINISHED

        return results

    async def _run_task_async(self, debug) -> None:
        """Runs the task asynchronously, awaiting the tasks that must be completed before this one can be run."""
        # logger.info(f"Running task for {self.question.question_name}")
        try:
            # This is waiting for the tasks that must be completed before this one can be run.
            # This does *not* use the return_exceptions = True flag, so if any of the tasks fail,
            # it throws the exception immediately, which is what we want.
            self.task_status = TaskStatus.WAITING_FOR_DEPENDENCIES
            # The 'self' here is a list of tasks that must be completed before this one can be run.
            await asyncio.gather(*self)
        except asyncio.CancelledError:
            self.status = TaskStatus.CANCELLED
            # logger.info(f"Task for {self.question.question_name} was cancelled, most likely because it was skipped.")
            raise
        except Exception as e:
            self.task_status = TaskStatus.PARENT_FAILED
            # breakpoint()
            # logger.error(f"Required tasks for {self.question.question_name} failed: {e}")
            # turns the parent exception into a custom exception
            # So the task gets canceled but this InterviewErrorPriorTaskCanceled exception
            # So we never get the question details we need.
            raise InterviewErrorPriorTaskCanceled(
                f"Required tasks failed for {self.question.question_name}"
            ) from e
        else:
            # logger.info(f"Tasks for {self.question.question_name} completed")
            # This is the actual task that we want to run.
            return await self._run_focal_task(debug)



from edsl.language_models.Anthropic import create_anthropic_model
from edsl.enums import LanguageModelType

model_name = LanguageModelType.ANTHROPIC_3_HAIKU.value

ClaudeHaiku = create_anthropic_model(
    model_name=model_name, model_class_name="ClaudeHaiku"
)

if __name__ == "__main__":
    model = ClaudeHaiku()
    from edsl import QuestionFreeText

    results = QuestionFreeText.example().by(model).run()
    results.select("answer.*", "model.*").print()



from edsl.jobs.buckets.TokenBucket import TokenBucket


class ModelBuckets:
    """A class to represent the token and request buckets for a model.

    Most LLM model services have limits both on requests-per-minute (RPM) and tokens-per-minute (TPM).
    A request is one call to the service. The number of tokens required for a request depends on parameters.
    """

    def __init__(self, requests_bucket: TokenBucket, tokens_bucket: TokenBucket):
        """Initialize the model buckets.

        The requests bucket captures requests per unit of time.
        The tokens bucket captures the number of language model tokens.

        """
        self.requests_bucket = requests_bucket
        self.tokens_bucket = tokens_bucket

    def __add__(self, other: "ModelBuckets"):
        """Combine two model buckets."""
        return ModelBuckets(
            requests_bucket=self.requests_bucket + other.requests_bucket,
            tokens_bucket=self.tokens_bucket + other.tokens_bucket,
        )

    @classmethod
    def infinity_bucket(cls, model_name: str = "not_specified") -> "ModelBuckets":
        """Create a bucket with infinite capacity and refill rate."""
        return cls(
            requests_bucket=TokenBucket(
                bucket_name=model_name,
                bucket_type="requests",
                capacity=float("inf"),
                refill_rate=float("inf"),
            ),
            tokens_bucket=TokenBucket(
                bucket_name=model_name,
                bucket_type="tokens",
                capacity=float("inf"),
                refill_rate=float("inf"),
            ),
        )

    def visualize(self):
        """Visualize the token and request buckets."""
        plot1 = self.requests_bucket.visualize()
        plot2 = self.tokens_bucket.visualize()
        return plot1, plot2

    def __repr__(self):
        return f"ModelBuckets(requests_bucket={self.requests_bucket}, tokens_bucket={self.tokens_bucket})"



"""Class for creating question instructions to be used in a survey."""
from edsl.prompts.Prompt import PromptBase
from edsl.prompts.prompt_config import ComponentTypes


class QuestionInstuctionsBase(PromptBase):
    """Class for creating question instructions to be used in a survey."""

    component_type = ComponentTypes.QUESTION_INSTRUCTIONS



import openai
import os
import re
from typing import Any
from openai import AsyncOpenAI
from edsl.enums import LanguageModelType, InferenceServiceType
from edsl.language_models import LanguageModel
from edsl.exceptions import MissingAPIKeyError

LanguageModelType.GPT_4.value


def create_openai_model(model_name, model_class_name) -> LanguageModel:
    class LLM(LanguageModel):
        """
        Child class of LanguageModel for interacting with OpenAI models
        """

        _inference_service_ = InferenceServiceType.OPENAI.value
        _model_ = model_name
        _parameters_ = {
            "temperature": 0.5,
            "max_tokens": 1000,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "logprobs": False,
            "top_logprobs": 3,
        }

        def get_headers(self) -> dict[str, Any]:
            from openai import OpenAI

            client = OpenAI()
            response = client.chat.completions.with_raw_response.create(
                messages=[
                    {
                        "role": "user",
                        "content": "Say this is a test",
                    }
                ],
                model=self.model,
            )
            return dict(response.headers)

        def get_rate_limits(self) -> dict[str, Any]:
            try:
                headers = self.get_headers()
            except Exception as e:
                return {
                    "rpm": 10_000,
                    "tpm": 2_000_000,
                }
            else:
                return {
                    "rpm": int(headers["x-ratelimit-limit-requests"]),
                    "tpm": int(headers["x-ratelimit-limit-tokens"]),
                }

        async def async_execute_model_call(
            self, user_prompt: str, system_prompt: str = ""
        ) -> dict[str, Any]:
            """Calls the OpenAI API and returns the API response."""
            self.client = AsyncOpenAI()
            response = await self.client.chat.completions.create(
                model=self.model,
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt},
                ],
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                logprobs=self.logprobs,
                top_logprobs=self.top_logprobs if self.logprobs else None,
            )
            return response.model_dump()

        @staticmethod
        def parse_response(raw_response: dict[str, Any]) -> str:
            """Parses the API response and returns the response text."""
            response = raw_response["choices"][0]["message"]["content"]
            pattern = r"^```json(?:\\n|\n)(.+?)(?:\\n|\n)```$"
            match = re.match(pattern, response, re.DOTALL)
            if match:
                return match.group(1)
            else:
                return response

    LLM.__name__ = model_class_name

    return LLM


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""Module for uploading and retrieving objects from a community server."""

import os
import requests
import json
import textwrap

from rich import print as print


SERVER_URL = os.getenv("EXPECTED_PARROT_URL")

from edsl.utilities.SystemInfo import SystemInfo


def post(
    object,
    *,
    username: str = None,
    description: str = None,
    system_info: str = None,
    public=True,
    ask=True,
):
    """Upload an object to the server with additional metadata."""
    info = SystemInfo("edsl")
    if ask:
        print(
            f"OK to send the folloing info to community server (info will be public)?"
        )
        print(info)
        inp = input("Enter y/n: ")
        if inp != "y":
            print("Canceled posting")
            return
    object_type = object.__class__.__name__
    object_data = json.dumps(object.to_dict()).encode("utf-8")
    # Prepare the metadata for the POST request
    username = username or info.username
    system_info = system_info or (info.system_info + ";" + info.release_info)
    edsl_version = info.package_version
    data = {
        "username": username,
        "system_info": system_info,
        "edsl_version": edsl_version,
        "object_type": object_type,
        "description": description,
        "public": str(public).lower(),  # Convert boolean to lowercase string
    }

    # Send the POST request with the object data and metadata
    response = requests.post(
        f"{SERVER_URL}/upload",
        files={"file": ("file", object_data, "application/json")},
        data=data,
    )

    # Process the response
    if response.status_code == 200:
        key = response.json()["key"]
        print(
            textwrap.dedent(
                f"""\
        EDSL object of type {object_type} uploaded successfully, with key `{key}`
        You can share this key with others to allow them to retrieve the object.
        To retrieve the object: 

        >>> from edsl import get
        >>> object = get('{key}')
        
        To view all community objects, go to {SERVER_URL}.
        To view this object, go to {SERVER_URL}/view/{key}.
        WARNING: This object is stored on a public server and can be accessed by anyone with the key.
        """
            )
        )
        return response.json()["key"]
    else:
        print(f"Error during file upload: {response.text}")
        return None


def get(key):
    """Retrieve an object and its type from the server using a key."""
    response = requests.get(f"{SERVER_URL}/retrieve/{key}")

    if response.status_code == 200:
        response_data = response.json()
        object_data = response_data.get(
            "object_data"
        )  # Assuming the object data is under "object_data"
        object_type = response_data.get("object_type")
        if object_type and object_data:
            object_dict = json.loads(object_data)

            from edsl import Agent
            from edsl.results import Results
            from edsl.jobs import Jobs
            from edsl.surveys import Survey
            from edsl.questions import Question

            mapping = {
                "Agent": Agent,
                "Question": Question,
                "Survey": Survey,
                "Results": Results,
                "Jobs": Jobs,
            }

            if object_type in mapping:
                object_type = mapping[object_type]
                object = object_type.from_dict(object_dict)
                return object
            else:
                print(f"Object type {object_type} is not supported.")
    else:
        print(f"Error during file retrieval: {response.text}")
        return None


def community():
    """Retrieve all objects from the server."""
    response = requests.get(f"{SERVER_URL}/community")
    if response.status_code == 200:
        return response.json()
    else:
        return f"Error: {response.status_code}"


if __name__ == "__main__":
    from edsl import Survey
    from edsl.results import Results

    msg = post(Survey.example())

    print(msg)
    new_obj = get(msg)
    print(new_obj)

    msg = post(Results.example())
    new_obj = get(msg)
    print(new_obj)
    new_obj.select("answer.*").print()



from edsl.jobs.interviews.InterviewStatusLog import InterviewStatusLog
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage

from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary


class InterviewStatusMixin:
    @property
    def has_exceptions(self) -> bool:
        """Return True if there are exceptions."""
        return len(self.exceptions) > 0

    @property
    def task_status_logs(self) -> InterviewStatusLog:
        """Return the task status logs for the interview.

        The keys are the question names; the values are the lists of status log changes for each task.
        """
        for task_creator in self.task_creators.values():
            self._task_status_log_dict[
                task_creator.question.question_name
            ] = task_creator.status_log
        return self._task_status_log_dict

    @property
    def token_usage(self) -> InterviewTokenUsage:
        """Determine how many tokens were used for the interview."""
        return self.task_creators.token_usage

    @property
    def interview_status(self) -> InterviewStatusDictionary:
        """Return a dictionary mapping task status codes to counts."""
        return self.task_creators.interview_status



# """This module contains the Result class, which captures the result of one interview."""
from __future__ import annotations
from collections import UserDict
from typing import Any, Type

from rich.table import Table

from IPython.display import display

from edsl.agents import Agent
from edsl.language_models import LanguageModel
from edsl.scenarios import Scenario

from edsl.utilities import is_notebook

from edsl.Base import Base

from collections import UserDict


class PromptDict(UserDict):
    """A dictionary that is used to store the prompt for a given result."""

    def rich_print(self):
        """Display an object as a table."""
        table = Table(title="")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self
        for attr_name, attr_value in to_display.items():
            table.add_row(attr_name, repr(attr_value))

        return table


def agent_namer_closure():
    """Return a function that can be used to name an agent."""
    agent_dict = {}

    def agent_namer(agent):
        """Return a name for an agent. If the agent has been named before, return the same name. Otherwise, return a new name."""
        nonlocal agent_dict
        agent_count = len(agent_dict)
        if id(agent) in agent_dict:
            return agent_dict[id(agent)]
        else:
            agent_dict[id(agent)] = f"Agent_{agent_count}"
            return agent_dict[id(agent)]

    return agent_namer


agent_namer = agent_namer_closure()


class Result(Base, UserDict):
    """
    This class captures the result of one interview.

    The answer dictionary has the structure:

    >>> Result.example().answer
    {'how_feeling': 'OK', 'how_feeling_comment': 'This is a real survey response from a human.', 'how_feeling_yesterday': 'Great', 'how_feeling_yesterday_comment': 'This is a real survey response from a human.'}

    Its main data is an Agent, a Scenario, a Model, an Iteration, and an Answer.
    These are stored both in the UserDict and as attributes.

    >>> results.select('question_text.how_feeling')
    >>> results.select('question_type.how_feeling')

    """

    def __init__(
        self,
        agent: Agent,
        scenario: Scenario,
        model: Type[LanguageModel],
        iteration: int,
        answer: str,
        prompt: dict[str, str] = None,
        raw_model_response=None,
        survey=None,
    ):
        """Initialize a Result object.

        :param agent: The Agent object.
        :param scenario: The Scenario object.
        :param model: The LanguageModel object.
        :param iteration: The iteration number.
        :param answer: The answer string.
        :param prompt: A dictionary of prompts.
        :param raw_model_response: The raw model response.

        """
        data = {
            "agent": agent,
            "scenario": scenario,
            "model": model,
            "iteration": iteration,
            "answer": answer,
            "prompt": prompt or {},
            "raw_model_response": raw_model_response or {},
        }
        super().__init__(**data)
        # but also store the data as attributes
        self.agent = agent
        self.scenario = scenario
        self.model = model
        self.iteration = iteration
        self.answer = answer
        self.prompt = prompt or {}
        self.raw_model_response = raw_model_response or {}
        self.survey = survey

        if survey is not None:
            self.question_to_attributes = {
                q.question_name: {
                    "question_text": q.question_text,
                    "question_type": q.question_type,
                }
                for q in survey.questions
            }
        else:
            self.question_to_attributes = {}

    ###############
    # Used in Results
    ###############
    @property
    def sub_dicts(self) -> dict[str, dict]:
        """Return a dictionary where keys are strings for each of the main class attributes/objects."""
        if self.agent.name is None:
            agent_name = agent_namer(self.agent)
        else:
            agent_name = self.agent.name

        question_text_dict = {}
        for key, _ in self.answer.items():
            if key in self.question_to_attributes:
                question_text_dict[
                    key + "_question_text"
                ] = self.question_to_attributes[key]["question_text"]

        return {
            "agent": self.agent.traits | {"agent_name": agent_name},
            "scenario": self.scenario,
            "model": self.model.parameters | {"model": self.model.model},
            "answer": self.answer,
            "prompt": self.prompt,
            "raw_model_response": self.raw_model_response,
            "iteration": {"iteration": self.iteration},
            "question_text": question_text_dict,
        }

    def code(self):
        """Return a string of code that can be used to recreate the Result object."""
        raise NotImplementedError

    @property
    def combined_dict(self) -> dict[str, Any]:
        """Return a dictionary that includes all sub_dicts, but also puts the key-value pairs in each sub_dict as a key_value pair in the combined dictionary."""
        combined = {}
        for key, sub_dict in self.sub_dicts.items():
            combined.update(sub_dict)
            combined.update({key: sub_dict})
        return combined

    def get_value(self, data_type: str, key: str) -> Any:
        """Return the value for a given data type and key.

        - data types can be "agent", "scenario", "model", or "answer"
        - keys are relevant attributes of the Objects the data types represent
        results.get_value("answer", "how_feeling") will return "Good" or "Bad" or whatnot
        """
        return self.sub_dicts[data_type][key]

    @property
    def key_to_data_type(self) -> dict[str, str]:
        """Return a dictionary where keys are object attributes and values are the data type (object) that the attribute is associated with."""
        d = {}
        for data_type in [
            "agent",
            "scenario",
            "model",
            "answer",
            "prompt",
            "raw_model_response",
            "iteration",
        ]:
            for key in self.sub_dicts[data_type]:
                d[key] = data_type
        return d

    def rows(self, index):
        """Return a generator of rows for the Result object."""
        for data_type, subdict in self.sub_dicts.items():
            for key, value in subdict.items():
                yield (index, data_type, key, str(value))

    ###############
    # Useful
    ###############
    def copy(self) -> Result:
        """Return a copy of the Result object."""
        return Result.from_dict(self.to_dict())

    def __eq__(self, other):
        """Return True if the Result object is equal to another Result object."""
        return self.to_dict() == other.to_dict()

    ###############
    # Serialization
    ###############
    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the Result object."""
        return {
            k: v if not hasattr(v, "to_dict") else v.to_dict() for k, v in self.items()
        }

    @classmethod
    def from_dict(self, json_dict: dict) -> Result:
        """Return a Result object from a dictionary representation."""
        result = Result(
            agent=Agent.from_dict(json_dict["agent"]),
            scenario=Scenario.from_dict(json_dict["scenario"]),
            model=LanguageModel.from_dict(json_dict["model"]),
            iteration=json_dict["iteration"],
            answer=json_dict["answer"],
            prompt=json_dict["prompt"],
            raw_model_response=json_dict.get(
                "raw_model_response", {"raw_model_response": "No raw model response"}
            ),
        )
        return result

    def rich_print(self):
        """Display an object as a table."""
        # from edsl.utilities import print_dict_with_rich
        from rich import print

        table = Table(title="Result")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self.__dict__.copy()
        data = to_display.pop("data", None)
        for attr_name, attr_value in to_display.items():
            if hasattr(attr_value, "rich_print"):
                table.add_row(attr_name, attr_value.rich_print())
            elif isinstance(attr_value, dict):
                a = PromptDict(attr_value)
                table.add_row(attr_name, a.rich_print())
            else:
                table.add_row(attr_name, repr(attr_value))
        return table

    def __repr__(self):
        """Return a string representation of the Result object."""
        return f"Result(agent={repr(self.agent)}, scenario={repr(self.scenario)}, model={repr(self.model)}, iteration={self.iteration}, answer={repr(self.answer)}, prompt={repr(self.prompt)}"

    @classmethod
    def example(cls):
        """Return an example Result object."""
        from edsl.results import Results

        return Results.example()[0]


def main():
    """Run the main function."""
    from edsl.results.Result import Result
    import json

    print("Being imported")
    json_string = """
    {
        "agent": {
            "traits": {
                "status": "Unhappy"
            }
        },
        "scenario": {
            "period": "morning"
        },
        "model": {
            "model": "gpt-3.5-turbo",
            "parameters": {
                "temperature": 0.5,
                "max_tokens": 1000,
                "top_p": 1,
                "frequency_penalty": 0,
                "presence_penalty": 0,
                "use_cache": true
            }
        },
        "iteration": 0,
        "answer": {
            "how_feeling": "Bad"
        }, 
        "prompt": {"how_feeling_user_prompt": "How are you feeling today?", "how_feeling_system_prompt": "Answer the question"}
    }
    """

    result = Result.from_dict(json.loads(json_string))

    result.sub_dicts
    assert result.combined_dict["how_feeling"] == "Bad"

    result.combined_dict
    assert result.get_value("answer", "how_feeling") == "Bad"

    result.key_to_data_type
    print(result)

    assert result == result.copy()

    result.to_dict()


if __name__ == "__main__":
    print(Result.example())



from rich.table import Table
from rich.console import Console
import math 

def logprob_to_prob(logprob):
    if logprob is None:
        return None
    return math.exp(logprob)


def format_output(data):
    content = data['choices'][0]['logprobs']['content']
    table = Table(show_header=True, header_style="bold magenta")

    # First pass to determine the maximum number of top tokens
    max_tokens = max(len(item['top_logprobs']) for item in content)
    
    # Set up the table columns
    table.add_column("Token", style="bold")
    for i in range(max_tokens):
        table.add_column(f"Top Token {i+1}")

    for item in content:
        token = item['token'].strip()
        top_tokens = [
            (top['token'].strip(), logprob_to_prob(top.get('logprob')))
            for top in item['top_logprobs']
        ]
        
        # Row data starts with the main token
        row = [token]

        # Add each top token with coloring based on probability
        for token, probability in top_tokens:
            if probability is not None:
                # Assign color based on probability value
                if probability > 0.5:
                    color = "green"
                elif probability > 0.2:
                    color = "yellow"
                else:
                    color = "red"
                formatted_token = f"[{color}]{token} ({probability:.2f})[/]"
            else:
                formatted_token = f"{token} (N/A)"
            row.append(formatted_token)
        
        # Ensure row has enough columns
        while len(row) < max_tokens + 1:
            row.append("")
        
        table.add_row(*row)

    console = Console()
    console.print(table)

class SimpleAskMixin:

    #def simple_ask(self, question: QuestionBase, system_prompt = "You are a helpful agent pretending to be a human.", top_logprobs = 2):

    def simple_ask(self, model = None, system_prompt = "You are a helpful agent pretending to be a human. Do not break character", top_logprobs = 4):
        from edsl import Model
        if model is None:
            model = Model()
        response = model.simple_ask(self, system_prompt, top_logprobs)
        return format_output(response)
        




"""Directed Acyclic Graph (DAG) class."""
from collections import UserDict
from graphlib import TopologicalSorter


class DAG(UserDict):
    """Class for creating a Directed Acyclic Graph (DAG) from a dictionary."""

    def __init__(self, data: dict):
        """Initialize the DAG class."""
        super().__init__(data)
        self.reverse_mapping = self._create_reverse_mapping()

    def _create_reverse_mapping(self):
        """
        Create a reverse mapping of the DAG, where the keys are the children and the values are the parents.

        Example usage:

        .. code-block:: python

            data = {"a": ["b", "c"], "b": ["d"], "c": [], "d": []}
            dag = DAG(data)
            dag._create_reverse_mapping()
            {'b': {'a'}, 'c': {'a'}, 'd': {'b'}}

        """
        rev_map = {}
        for key, values in self.items():
            for value in values:
                rev_map.setdefault(value, set()).add(key)
        return rev_map

    def get_all_children(self, key):
        """Get all children of a node in the DAG."""
        children = set()

        def dfs(node):
            for child in self.reverse_mapping.get(node, []):
                if child not in children:
                    children.add(child)
                    dfs(child)

        dfs(key)
        return children

    def topologically_sorted_nodes(self):
        """
        Return a sequence of the DAG.

        Example usage:

        .. code-block:: python

            data = {"a": ["b", "c"], "b": ["d"], "c": [], "d": []}
            dag = DAG(data)
            dag.topologically_sorted_nodes() == ['c', 'd', 'b', 'a']
            True

        """
        return list(TopologicalSorter(self).static_order())

    def __add__(self, other_dag):
        """Combine two DAGs."""
        d = {}
        combined_keys = set(self.keys()).union(set(other_dag.keys()))
        for key in combined_keys:
            d[key] = self.get(key, set({})).union(other_dag.get(key, set({})))
        return DAG(d)
        # if textify:
        #     return DAG(self.textify(d))
        # else:
        #     return DAG(d)

    @classmethod
    def example(cls):
        """Return an example of the `DAG`."""
        data = {"a": ["b", "c"], "b": ["d"], "c": [], "d": []}
        return cls(data)


if __name__ == "__main__":
    import doctest

    doctest.testmod()
